//
// Copyright (c) 1997 X/Open Company Ltd., A member of The Open Group.
//
// All rights reserved. No part of this source code may be reproduced,
// stored in a retrieval system, or transmitted, in any form or by any
// means, electronic, mechanical, photocopying, recording or otherwise,
// except as stated in the end-user licence agreement, without the prior
// permission of the copyright owners.
//
// Developed for X/Open by ApTest Ireland Limited.
//
// Project: VSORB
//
// File: src/utils/nodes/src/node.cpp
//
// Modifications:
// $Log: node.cpp,v $
// Revision 1.2  2000/02/11 16:05:20  mli
// Decl. of index i extracted out of the 'for' loop (10x).
//
// Revision 1.3  1998/05/11 10:52:30  stephen
// Operator set for Node types updated.
//
// Revision 1.2  1997/10/16 14:23:43  james
// Reduced memory usage
//
// Revision 1.1  1997/10/06 11:36:02  james
// Initial Sources
//

#include <stdio.h>
#include <stdlib.h>
#include <memory.h>
#include <string.h>

#include "node.hh"

Node notInitialised;

// Member functions for the class Node
// This is the base class from which all nodes must be derived

Node::Node()
{
  parent = 0;
  nodeId = IdNotInitialised;
#ifdef GENERATOR_FUNCTIONS
  memset(recursionCount,0,sizeof(RecursionCount));
#endif
}

Node::Node(NodeId n)
{
  parent = 0;
  nodeId = n;
#ifdef GENERATOR_FUNCTIONS
  memset(recursionCount,0,sizeof(RecursionCount));
#endif
}

Node::Node(Node *p, NodeId n, RecursionCount r)
{
  parent = p;
  nodeId = n;
#ifdef GENERATOR_FUNCTIONS
  if (r == 0)
    memset(recursionCount,0,sizeof(RecursionCount));
  else
    memcpy(recursionCount,r,sizeof(RecursionCount));

  // Update the RecursionCount array to note that an instance of this class
  // has been created
  ++recursionCount[nodeId];
#endif
}

Node::Node(const Node &r)
{
  parent = r.parent;
  nodeId = r.nodeId;
#ifdef GENERATOR_FUNCTIONS
  memcpy(recursionCount,r.recursionCount,sizeof(RecursionCount));
#endif
}

Node::~Node()
{
}

Node &Node::operator=(const Node &r)
{
  if (&r != this)
  {
    if (nodeId != IdNotInitialised && nodeId != r.nodeId)
    {
      cerr << "Error: Cannot assign " << r.nodeId << " to " << nodeId << endl;
      exit(EXIT_FAILURE);
    }
    nodeId = r.nodeId;
    //parent and recursionCount are not assigned here, if necessary these are
    // assigned during construction
  }

  return *this;
}

NodeId Node::Id() const
{
  return nodeId;
}

Node::operator NodeId() const
{
    return nodeId;
}

Node *Node::Parent() const
{
  return parent;
}

void Node::Parent(Node *p)
{
  parent = p;
}

int Node::Choice() const
{
  cerr  << "Error: Choice() called on " << nodeId << endl;
  exit(EXIT_FAILURE);
  return 0;
}



Node &Node::operator[](int )
{
  return notInitialised;
}

Node &Node::operator[](NodeId ) 
{
  return notInitialised;
}
Node &Node::operator[](NodeId ) const
{
  return notInitialised;
}

int Node::IsEmpty()
{
  return 1;
}

int Node::Length()
{
  return 0;
}


ExprValue Node::exprValue()
{
  ExprValue ret;
  return ret;
}



// Member functions for the SimpleChoice Node
// This node is used as a base class for node that represent a choice between
// two or more terminals
// eg <floating_pt_type> ::= "float" | "double" | "long" "double"
// A class derived from this class must provide its own constructor in order
// to initialise maxChoice and the Output functions in order to output the
// correct terminal based on the value of choice.

SimpleChoiceNode::SimpleChoiceNode() : Node()
{
  choice = 0;
  maxChoice = 0;
}

SimpleChoiceNode::SimpleChoiceNode(NodeId n) : Node(n)
{
  choice = 0;
  maxChoice = 0;
}

SimpleChoiceNode::SimpleChoiceNode(Node *p, NodeId n, RecursionCount r) : Node(p,n,r)
{
  choice = 0;
  maxChoice = 0;
}

SimpleChoiceNode::SimpleChoiceNode(const SimpleChoiceNode &r)
{
  parent = r.parent;
  nodeId = r.nodeId;
#ifdef GENERATOR_FUNCTIONS
  memcpy(recursionCount,r.recursionCount,sizeof(RecursionCount));
#endif

  choice = r.choice;
  maxChoice = r.maxChoice;
}

SimpleChoiceNode::~SimpleChoiceNode()
{
}

int SimpleChoiceNode::Choice() const
{
  return choice;
}

Node &SimpleChoiceNode::operator=(const Node &r)
{
  if (&r != this)
  {
    Node::operator=(r);
  
    choice = ((SimpleChoiceNode &)r).choice;
    maxChoice = ((SimpleChoiceNode &)r).maxChoice;
  }

  return *this;
}

Node &SimpleChoiceNode::operator=(const int r)
{
  if (choice > maxChoice)
  {
    cerr  << "Error: maxChoice exceeded in assignment operator for " 
          << nodeId << endl;
    exit(EXIT_FAILURE);
  }

  choice = r;

  return *this;
}



// Member functions for the Choice Node
// This node is used as a base class for node that represent a choice between
// two or more non-terminals
// eg <interface> ::= <interface_dcl> | <forward_dcl>
// A class derived from this class must provide its own constructor in order
// to initialise childIds with the list of possible child nodes
// It may also provide the Output functions. The default behaviour is to 
// call the child's output function.

ChoiceNode::ChoiceNode() : Node()
{
  choice = 0;
  memset(childIds,0,sizeof(childIds));
  child = 0;
}

ChoiceNode::ChoiceNode(NodeId n) : Node(n)
{
  choice = 0;
  memset(childIds,0,sizeof(childIds));
  child = 0;
}

ChoiceNode::ChoiceNode(Node *p, NodeId n, RecursionCount r) : Node(p,n,r)
{
  choice = 0;
  memset(childIds,0,sizeof(childIds));
  child = 0;
}

ChoiceNode::ChoiceNode(const ChoiceNode &r)
{
  parent = r.parent;
  nodeId = r.nodeId;
#ifdef GENERATOR_FUNCTIONS
  memcpy(recursionCount,r.recursionCount,sizeof(RecursionCount));
#endif

  choice = r.choice;
  memcpy(childIds,r.childIds,sizeof(childIds));
#ifdef GENERATOR_FUNCTIONS
  child = NewNode(this,childIds[choice],recursionCount);
#else
  child = NewNode(this,childIds[choice]);
#endif
  *child = *r.child;  // Deep Copy
}

ChoiceNode::~ChoiceNode()
{
  // Delete the child, if it exists
  if (child != 0)
    delete child;
}

int ChoiceNode::Choice() const
{
  return choice;
}

Node &ChoiceNode::operator=(const Node &r)
{
  if (&r != this)
  {
    Node::operator=(r);
  
    choice = ((ChoiceNode &)r).choice;
    memcpy(childIds, ((ChoiceNode &)r).childIds,sizeof(childIds));

    if (child != 0)
    {
      delete child;
      child = 0;
    }

    if (((ChoiceNode &)r).child != 0)
    {
#ifdef GENERATOR_FUNCTIONS
      child = NewNode(this,childIds[choice],recursionCount);
#else
      child = NewNode(this,childIds[choice]);
#endif
      *child = *(((ChoiceNode &)r).child);
    }
  }

  return *this;
}

ChoiceNode &ChoiceNode::operator<<(Node *r)
{
  if (r != 0)
  {
    // Verify that node is of the correct type

    int i;
    
    for (i = 0; childIds[i] != 0 && r->Id() != childIds[i]; ++i);
    if (childIds[i] == 0)
    {
      cerr << "Error: " << nodeId << " cannot contain " << r->Id() << endl;
      exit(EXIT_FAILURE);
    }
  
    if (child)
      delete(child);
  
    choice = i;
    child = r;
    child->Parent(this);
  }

  return *this;
}

ChoiceNode &ChoiceNode::operator<<(Node &r)
{
  // Verify that node is of the correct type
  int i;
  for (i = 0; childIds[i] != 0 && r.Id() != childIds[i]; ++i);
  if (childIds[i] == 0)
  {
    cerr << "Error: " << nodeId << " cannot contain " << r.Id() << endl;
    exit(EXIT_FAILURE);
  }

  if (child)
    delete(child);

  choice = i;
#ifdef GENERATOR_FUNCTIONS
  child = NewNode(this,childIds[choice],recursionCount);
#else
  child = NewNode(this,childIds[choice]);
#endif
  *child = r;  // Deep-copy assignment

  return *this;
}



Node &ChoiceNode::operator[](int i)
{
  return (i == 0 && child != 0)?*child:notInitialised;
}

Node &ChoiceNode::operator[](NodeId id) 
{
  return (childIds[choice] == id && child != 0)?*child:notInitialised;
}

Node &ChoiceNode::operator[](NodeId id) const
{
  return (childIds[choice] == id && child != 0)?*child:notInitialised;
}

int ChoiceNode::IsEmpty()
{
  return (child == 0);
}

int ChoiceNode::Length()
{
  return (child != 0)?1:0;
}


ExprValue ChoiceNode::exprValue()
{
  ExprValue v;
   
  if (child != 0)
    v = child->exprValue();

  return v;
}







// Member functions for the Sequence Node
// This node is used as a base class for node that represent a sequence of 
// two or more non-terminals
// eg <type_declarator> ::= <type_spec> | <declarators>
// A class derived from this class must provide its own constructor in order
// to initialise childIds with the list of child nodes
// It may also provide the Output functions. The default behaviour is to 
// call the each of the children's output functions in turn.

SequenceNode::SequenceNode() : Node()
{
  memset(childIds,0,sizeof(childIds));
  memset(childIsOptional,0,sizeof(childIsOptional));
  memset(childNodes,0,sizeof(childNodes));
}

SequenceNode::SequenceNode(NodeId n) : Node(n)
{
  memset(childIds,0,sizeof(childIds));
  memset(childIsOptional,0,sizeof(childIsOptional));
  memset(childNodes,0,sizeof(childNodes));
}

SequenceNode::SequenceNode(Node *p, NodeId n, RecursionCount r) : Node(p,n,r)
{
  memset(childIds,0,sizeof(childIds));
  memset(childIsOptional,0,sizeof(childIsOptional));
  memset(childNodes,0,sizeof(childNodes));
}

SequenceNode::SequenceNode(const SequenceNode &r)
{
  parent = r.parent;
  nodeId = r.nodeId;
#ifdef GENERATOR_FUNCTIONS
  memcpy(recursionCount,r.recursionCount,sizeof(RecursionCount));
#endif

  memcpy(childIds,r.childIds,sizeof(childIds));
  memcpy(childIsOptional,r.childIsOptional,sizeof(childIsOptional));
  for(int i = 0; childIds[i] != 0; ++i)
  {
    if (r.childNodes[i] != 0)
    {
#ifdef GENERATOR_FUNCTIONS
      childNodes[i] = NewNode(this,childIds[i],recursionCount);
#else
      childNodes[i] = NewNode(this,childIds[i]);
#endif
      *childNodes[i] = *r.childNodes[i]; // Deep Copy
    }
  }
}

SequenceNode::~SequenceNode()
{
  // Delete all child nodes
  for(int i = 0; childIds[i] != 0; ++i)
  {
    if (childNodes[i] != 0)
      delete childNodes[i];
  }
}

Node &SequenceNode::operator=(const Node &r)
{
  if (&r != this)
  {
    Node::operator=(r);
  
    memcpy(childIds, ((SequenceNode &)r).childIds,sizeof(childIds));
    memcpy(childIsOptional, ((SequenceNode &)r).childIsOptional,sizeof(childIsOptional));
  
    for(int i = 0; childIds[i] != 0; ++i)
    {
      if (((SequenceNode &)r).childNodes[i] != 0)
      {
#ifdef GENERATOR_FUNCTIONS
        childNodes[i] = NewNode(this,childIds[i],recursionCount);
#else
        childNodes[i] = NewNode(this,childIds[i]);
#endif
        *childNodes[i] = *(((SequenceNode &)r).childNodes[i]);
      }
    }
  }

  return *this;
}

SequenceNode &SequenceNode::operator<<(Node *r)
{
  if (r != 0)
  {
    // Verify that node is of the correct type
    int i;
    
    for (i = 0; childIds[i] != 0 && r->Id() != childIds[i]; ++i);
    if (childIds[i] == 0)
    {
      cerr << "Error: " << nodeId << " cannot contain " << r->Id() << endl;
      exit(EXIT_FAILURE);
    }
  
    if (childNodes[i])
      delete(childNodes[i]);
  
    childNodes[i] = r;
    childNodes[i]->Parent(this);
  }

  return *this;
}


SequenceNode &SequenceNode::operator<<(Node &r)
{
  // Verify that node is of the correct type
  int i;
    
  for (i = 0; childIds[i] != 0 && r.Id() != childIds[i]; ++i);
  if (childIds[i] == 0)
  {
    cerr << "Error: " << nodeId << " cannot contain " << r.Id() << endl;
    exit(EXIT_FAILURE);
  }

  if (childNodes[i])
    delete(childNodes[i]);

#ifdef GENERATOR_FUNCTIONS
  childNodes[i] = NewNode(this,childIds[i],recursionCount);
#else
  childNodes[i] = NewNode(this,childIds[i]);
#endif
  *childNodes[i] = r;  // Deep-copy assignment

  return *this;
}



Node &SequenceNode::operator[](int i)
{
  return (i >= 0 && i < Length() && childNodes[i] != 0)?*childNodes[i]:notInitialised;
}

Node &SequenceNode::operator[](NodeId id) 
{
  // Find child of appropiate type
  int i;
  for(i = 0; childIds[i] != 0 && childIds[i] != id; ++i);

  return (childNodes[i] != 0)?*childNodes[i]:notInitialised;
}

Node &SequenceNode::operator[](NodeId id) const
{
  // Find child of appropiate type
  int i;
  for(i = 0; childIds[i] != 0 && childIds[i] != id; ++i);

  return (childNodes[i] != 0)?*childNodes[i]:notInitialised;
}

int SequenceNode::IsEmpty()
{
  return (childNodes[0] == 0);
}

int SequenceNode::Length()
{
  int i;
  for(i = 0; childIds[i] != 0; ++i);
  
  return i;
}


// Member functions for the List Node
// This node is used as a base class for node that represent a List of 
// one or more non-terminals
// eg <specification> ::= <definition>+
// A class derived from this class must provide its own constructor in order
// to initialise childId with the id of the child nodes
// It may also provide the Output functions. The default behaviour is to 
// call the each of the children's output functions in turn.

ListNode::ListNode() : Node()
{
  childId = IdNotInitialised;
  separator = 0;
	maxChildNodes = 10;
	childNodes = new Node *[maxChildNodes+1];
  memset(childNodes,0,sizeof(Node *)*(maxChildNodes+1));
}

ListNode::ListNode(NodeId n) : Node(n)
{
  childId = IdNotInitialised;
  separator = 0;
	maxChildNodes = 10;
	childNodes = new Node *[maxChildNodes+1];
  memset(childNodes,0,sizeof(Node *)*(maxChildNodes+1));
}

ListNode::ListNode(Node *p, NodeId n, RecursionCount r) : Node(p,n,r)
{
  childId = IdNotInitialised;
  separator = 0;
	maxChildNodes = 10;
	childNodes = new Node *[maxChildNodes+1];
  memset(childNodes,0,sizeof(Node *)*(maxChildNodes+1));
}

ListNode::ListNode(const ListNode &r)
{
  parent = r.parent;
  nodeId = r.nodeId;
#ifdef GENERATOR_FUNCTIONS
  memcpy(recursionCount,r.recursionCount,sizeof(RecursionCount));
#endif
	maxChildNodes = r.maxChildNodes;
	childNodes = new Node *[maxChildNodes+1];
  memset(childNodes,0,sizeof(Node *)*(maxChildNodes+1));

  childId = r.childId;
  separator = r.separator;
  for(int i = 0; r.childNodes[i] != 0; ++i)
  {
#ifdef GENERATOR_FUNCTIONS
    childNodes[i] = NewNode(this,childId,recursionCount);
#else
    childNodes[i] = NewNode(this,childId);
#endif
    *childNodes[i] = *r.childNodes[i];
  }
}

ListNode::~ListNode()
{
  // Delete all the children
  for(int i = 0; childNodes[i] != 0; ++i)
    delete childNodes[i];
	delete childNodes;
}


Node &ListNode::operator=(const Node &r)
{
	int	i;

  if (&r != this)
  {
    Node::operator=(r);
  
		
  	for(i = 0; childNodes[i] != 0; ++i)
		{
    	delete childNodes[i];
			childNodes[i] = 0;
		}
		delete childNodes;

    childId = ((ListNode &)r).childId;
    separator = ((ListNode &)r).separator;
		maxChildNodes = ((ListNode &)r).maxChildNodes;
		childNodes = new Node *[maxChildNodes+1];
  	memset(childNodes,0,sizeof(Node *)*(maxChildNodes+1));
  
    for(i = 0; ((ListNode &)r).childNodes[i] != 0; ++i)
    {
#ifdef GENERATOR_FUNCTIONS
      childNodes[i] = NewNode(this,childId,recursionCount);
#else
      childNodes[i] = NewNode(this,childId);
#endif
      *childNodes[i] = *(((ListNode &)r).childNodes[i]);
    }
  }

  return *this;
}

ListNode &ListNode::operator<<(Node *r)
{
  if (r != 0)
  {
    // Verify that node is of the correct type
    if (r->Id() != childId)
    {
      cerr << "Error: " << nodeId << " cannot contain " << r->Id() << endl;
      exit(EXIT_FAILURE);
    }

    int i;
    for (i = 0; i < maxChildNodes && childNodes[i] != 0; ++i);
  
    if (i == maxChildNodes)
    {
			Node **tmp;

			tmp = new Node *[maxChildNodes*2+1];
  		memset(tmp,0,sizeof(Node *)*(maxChildNodes*2+1));
			memcpy(tmp,childNodes,sizeof(Node *)*maxChildNodes);
			delete childNodes;
			childNodes = tmp;
			maxChildNodes *=2;
    }
  
    childNodes[i] = r;
    childNodes[i]->Parent(this);
  }

  return *this;
}

ListNode &ListNode::operator<<(Node &r)
{
  // Verify that node is of the correct type
  if (r.Id() != childId)
  {
    cerr << "Error: " << nodeId << " cannot contain " << r.Id() << endl;
    exit(EXIT_FAILURE);
  }
    
  int i;
  for (i = 0; i < maxChildNodes && childNodes[i] != 0; ++i);

   if (i == maxChildNodes)
   {
		Node **tmp;

		tmp = new Node *[maxChildNodes*2+1];
  	memset(tmp,0,sizeof(Node *)*(maxChildNodes*2+1));
		memcpy(tmp,childNodes,sizeof(Node *)*maxChildNodes);
		delete childNodes;
		childNodes = tmp;
		maxChildNodes *=2;
  }

#ifdef GENERATOR_FUNCTIONS
  childNodes[i] = NewNode(this,childId,recursionCount);
#else
  childNodes[i] = NewNode(this,childId);
#endif
  *childNodes[i] = r;  // Deep-copy assignment

  return *this;
}

Node &ListNode::operator[](int i)
{
  return (i >= 0 && i < Length() && childNodes[i] != 0)?*childNodes[i]:notInitialised;
}

Node &ListNode::operator[](NodeId ) 
{
  // All children are of the same type, so this operator does not work for list nodes
  return notInitialised;
}

Node &ListNode::operator[](NodeId ) const
{
  // All children are of the same type, so this operator does not work for list nodes
  return notInitialised;
}

int ListNode::IsEmpty()
{
  return (childNodes[0] == 0);
}

int ListNode::Length()
{
  int i;
  for(i = 0; childNodes[i] != 0; ++i);
  
  return i;
}

char *ListNode::Separator()
{
	return separator;
}


